using Autofac;
using FitConnect.Exceptions;
using FitConnect.Interfaces;
using FitConnect.Interfaces.Subscriber;
using Microsoft.Extensions.Logging;
using Microsoft.IdentityModel.Logging;
using Newtonsoft.Json.Linq;

namespace FitConnect;

public static class ClientFactory {
    static ClientFactory() {
#if DEBUG
        IdentityModelEventSource.ShowPII = true;
#endif
    }

    /// <summary>
    ///     Creates the Sender Client
    /// </summary>
    /// <param name="environment">FIT-Connect Environment</param>
    /// <param name="clientId">Client ID</param>
    /// <param name="clientSecret">Client Secret</param>
    /// <param name="logger">An instance of an ILogger</param>
    /// <returns></returns>
    public static ISender GetSenderClient(FitConnectEnvironment environment, string clientId,
        string clientSecret, ILogger? logger = null) {
        try {
            return new Sender(environment, clientId, clientSecret, logger);
        }
        catch (Exception e) {
            throw new FitConnectInitialisationException("Failed to initialise sender", e);
        }
    }

    /// <summary>
    ///     Creates a subscriber client
    /// </summary>
    /// <param name="environment">FIT-Connect Environment</param>
    /// <param name="clientId">Client ID</param>
    /// <param name="clientSecret">Client Secret</param>
    /// <param name="privateDecryptionKeys">Private key for decryption</param>
    /// <param name="privateSigningKey">Private key for signing</param>
    /// <param name="logger">An instance of an ILogger</param>
    /// <returns></returns>
    public static ISubscriber GetSubscriberClient(FitConnectEnvironment environment,
        string clientId, string clientSecret,
        List<string> privateDecryptionKeys,
        string privateSigningKey,
        ILogger? logger = null) {

        CheckKeysForValidJsonFormat(privateDecryptionKeys, privateSigningKey);

        if (privateDecryptionKeys.Count != 1)
            throw new FitConnectInitialisationException(
                "Only one private key for decryption is supported");

        try {
            return new Subscriber(environment, clientId, clientSecret, privateDecryptionKeys,
                privateSigningKey, logger);
        }
        catch (Exception e) {
            throw new FitConnectInitialisationException("Failed to initialise subscriber", e);
        }
    }


    private static void CheckKeysForValidJsonFormat(List<string> privateDecryptionKeys,
        string privateSigningKey) {
        foreach (var key in privateDecryptionKeys) {
            try {
                var json = JObject.Parse(key);
            }
            catch (Exception e) {
                throw new FitConnectInitialisationException(
                    "Private decryption key is not in valid JSON format", e);
            }
        }

        try {
            var json = JObject.Parse(privateSigningKey);
        }
        catch (Exception e) {
            throw new FitConnectInitialisationException(
                "Private signing key is not in valid JSON format", e);
        }
    }

    /// <summary>
    ///     Returns a Router API implementation
    /// </summary>
    /// <param name="environment">FIT-Connect Environment</param>
    /// <param name="logger">An instance of an ILogger</param>
    /// <returns></returns>
    public static IRouter
        GetRouterClient(FitConnectEnvironment environment, ILogger? logger = null) {
        return new Router(environment, logger);
    }

#if DEBUG
    public static ISender GetSenderClient(IContainer container) {
        var settings = container.Resolve<IFitConnectSettings>();

        return new Sender(FitConnectEnvironment.Test, settings.SenderClientId,
            settings.SenderClientSecret, container.Resolve<ILogger>());
    }

    /// <summary>
    ///     Just for testing
    /// </summary>
    /// <returns></returns>
    internal static ISubscriber GetSubscriberClient(IContainer container) {
        return new Subscriber(container);
    }

#endif
}
