using System.IdentityModel.Tokens.Jwt;
using System.Security;
using System.Text;
using FitConnect.Models;
using FitConnect.Models.Api.Metadata;
using FitConnect.Models.v1.Api;
using IdentityModel;
using Microsoft.Extensions.Logging;
using Microsoft.IdentityModel.JsonWebTokens;
using Microsoft.IdentityModel.Tokens;
using Newtonsoft.Json;

namespace FitConnect.Encryption;

public class KeySet {
    public string? PrivateKeyDecryption { get; set; }
    public string? PrivateKeySigning { get; set; }
    public string? PublicKeyEncryption { get; set; }
    public string? PublicKeySignatureVerification { get; set; }
}

public class FitEncryption {
    private readonly IEncryptor _encryptor = new JoseEncryptor();
    private readonly ILogger? _logger;
    private JwtHeader? _jwtHeader;

    internal FitEncryption(ILogger? logger) {
        _logger = logger;
    }

    public FitEncryption(List<string>? privateDecryptionKeys, string? privateSigningKey,
        ILogger? logger) :
        this(logger) {
        PrivateDecryptionKeys = privateDecryptionKeys;
        PrivateSigningKey = privateSigningKey;
    }

    public FitEncryption(
        KeySet keySet,
        ILogger? logger) : this(logger) {
        PrivateDecryptionKeys = keySet.PrivateKeyDecryption != null
            ? new List<string>() { keySet.PrivateKeyDecryption }
            : null;
        PrivateSigningKey = keySet.PrivateKeySigning;
        PublicEncryptionKey = keySet.PublicKeyEncryption;
        PublicSignatureVerificationKey = keySet.PublicKeySignatureVerification;
    }

    public FitEncryption(string publicEncryptionKey, ILogger? logger) : this(logger) {
        PublicEncryptionKey = publicEncryptionKey;
    }

    public List<string>? PrivateDecryptionKeys { get; set; }
    public string? PrivateSigningKey { get; set; }
    public string? PublicEncryptionKey { get; set; }
    public string? PublicSignatureVerificationKey { get; set; }


    public (string plainText, byte[] plainBytes, byte[] tag) Decrypt(string cypherText,
        List<string> key) => _encryptor.Decrypt(key, cypherText);

    public (string plainText, byte[] plainBytes, byte[] tag) Decrypt(string cypherText) {
        if (PrivateDecryptionKeys == null)
            throw new InvalidOperationException("PrivateKey is not provided");

        return Decrypt(cypherText, PrivateDecryptionKeys);
    }


    public string Encrypt(string plainText, string key) {
        return _encryptor.Encrypt(key, plainText);
    }

    public string Encrypt(string plain) {
        if (PublicEncryptionKey == null)
            throw new InvalidOperationException("PublicKey is not provided");

        return Encrypt(plain, PublicEncryptionKey);
    }

    public string Encrypt(byte[] plain, string key) {
        return _encryptor.Encrypt(key, plain);
    }

    public string Encrypt(byte[] plain) {
        if (PublicEncryptionKey == null)
            throw new InvalidOperationException("PublicKey is not provided");

        return Encrypt(plain, PublicEncryptionKey);
    }

    public async Task<string> CreateRejectSecurityEventToken(Guid submissionId,
        Guid caseId,
        Guid destinationId, Problems[] problemsArray) {
        return await CreateSecurityEventToken(submissionId, caseId, destinationId,
            new Events {
                RejectSubmissionEvent = new RejectSubmissionEvent(problemsArray)
            }
        );
    }

    internal static AuthenticationTags GenerateAuthenticationTags(Submission submission) {
        var attachmentDictionary = new Dictionary<Guid, string>();
        submission.Attachments?.ForEach(a =>
            attachmentDictionary.Add(a.Id, a.AttachmentAuthentication!));

        return new AuthenticationTags {
            Data = submission.DataAuthentication,
            Metadata = submission.MetaAuthentication,
            Attachments = attachmentDictionary
        };
    }


    public async Task<string> CreateAcceptSecurityEventToken(ISubmissionBody submission,
        IEnumerable<Problems>? problems = null,
        dynamic? payload = null) {
        if (submission?.SubmissionId == null || submission?.CaseId == null ||
            submission?.DestinationId == null)
            throw new ArgumentException("SubmissionId, CaseId and DestinationId are required");

        return await CreateSecurityEventToken(submission.SubmissionId, submission.CaseId,
            submission.DestinationId,
            new Events {
                AcceptSubmissionEvent = new AcceptSubmissionEvent(payload, problems)
            });
    }


    public async Task<string> CreateAcceptSecurityEventToken(Submission submission,
        IEnumerable<Problems>? problems) {
        var payload = GenerateAuthenticationTags(submission);
        return await CreateSecurityEventToken(submission.Id, submission.CaseId,
            submission.DestinationId,
            new Events {
                AcceptSubmissionEvent = new AcceptSubmissionEvent(payload, problems)
            });
    }


    internal async Task<string> CreateSecurityEventToken(Guid submissionId,
        Guid caseId,
        Guid destinationId,
        Events submissionEvent) {
        var signingKey = new JsonWebKey(PrivateSigningKey);
        var transactionId = "case:" + caseId;
        var subject = "submission:" + submissionId;

        _jwtHeader =
            new JwtHeader(new SigningCredentials(signingKey, EncryptionSettings.SigningAlgorithm));

        JsonExtensions.Serializer = JsonConvert.SerializeObject;
        JsonExtensions.Deserializer = JsonConvert.DeserializeObject;

        _jwtHeader["typ"] = EncryptionSettings.JwtHeaderType;

        var jwtPayload = new JwtPayload {
            { "sub", subject },
            { "jti", Guid.NewGuid().ToString() },
            { "iat", DateTime.UtcNow.ToEpochTime() },
            { "iss", destinationId },
            { "txn", transactionId },
            { "events", submissionEvent }
        };

        jwtPayload.Add("$schema", ProjectSpecification.SetJsonSchemaUrl);

        var token = new JwtSecurityToken(
            _jwtHeader, jwtPayload) {
            SigningKey = signingKey
        };

        var handler = new JwtSecurityTokenHandler();
        var tokenString = handler.WriteToken(token);

        {
            var jsonData = Base64UrlEncoder.Decode(tokenString.Split('.')[1]);
            if (!await JsonHelper.ValidateJsonSchema(
                    new Uri(ProjectSpecification.SetJsonSchemaUrl),
                    jsonData
                ))
                throw new Exception("Invalid token schema for \n" + jsonData);
        }

        return tokenString;
    }

    public static DataHash CalculateHash(string data) {
        return new DataHash {
            Type = EncryptionSettings.ShaType,
            Content = ByteToHexString(
                EncryptionSettings.CalculateHash(Encoding.UTF8.GetBytes(data)))
        };
    }

    public static DataHash CalculateHash(byte[] data) {
        return new DataHash {
            Type = EncryptionSettings.ShaType,
            Content = ByteToHexString(EncryptionSettings.CalculateHash(data))
        };
    }

    private static string ByteToHexString(IEnumerable<byte> data) {
        var sb = new StringBuilder();
        foreach (var b in data) sb.Append(b.ToString("x2"));

        return sb.ToString();
    }


    public static bool VerifyJwt(string signature, IEnumerable<JsonWebKey> keys,
        ILogger? logger = null) {
        foreach (var key in keys)
            try {
                var valid = VerifyJwt(signature, key, logger);
                if (valid) return true;
            }
            catch (Exception e) {
                logger?.LogError(e, "Error verifying JWT");
            }

        return false;
    }

    public static bool VerifyJwt(string signature, string secret,
        ILogger? logger = null) {
        return VerifyJwt(signature,
            new JsonWebKey(secret), logger);
    }

    public static bool VerifyJwt(string signature, JsonWebKey key,
        ILogger? logger = null) {
        var tokenHandler = new JsonWebTokenHandler();

        logger?.LogTrace("Verifying JWT {Token}", signature);

        if (key.KeySize != EncryptionSettings.MinKeySize)
            throw new SecurityException($"Key size must be {EncryptionSettings.MinKeySize}");

        var result = tokenHandler.ValidateToken(signature, new TokenValidationParameters {
            ValidAlgorithms = EncryptionSettings.ValidTokenAlgorithms,
            IssuerSigningKey = key,
            ValidateIssuerSigningKey = true,
            ValidateIssuer = false,
            ValidateAudience = false,
            ValidateLifetime = true,
            ValidateActor = true,
            ValidateSignatureLast = true,
            ValidateTokenReplay = true,
            ValidateWithLKG = true,
            RequireAudience = true,
            RefreshBeforeValidation = true,
            RequireSignedTokens = true,
            RequireExpirationTime = false,
            IgnoreTrailingSlashWhenValidatingAudience = true,
            TryAllIssuerSigningKeys = true
        });

        var error = result.Exception?.Message;

        if (error != null)
            logger?.LogWarning("Validation error {Error}", error);
        else
            logger?.LogInformation("JWT is valid");

        return result.IsValid;
    }
}
